#include "BrainFunk.h"

#include "ui_BrainFunk.h"

#include "Machine.h"
#include "IOHandler.h"
#include "TapeModel.h"
#include "ProgramHighlighter.h"

#include "Challenges.h"
#include "ChallengeModel.h"

#include <QMenu>
#include <QMessageBox>
#include <QInputDialog>
#include <QFileDialog>
#include <QFont>

#include <QFile>
#include <QTextStream>

#include <QCloseEvent>

#include <QDebug>

enum Step
{
  Step1,
  Step10,
  Step100,
  Step1000,
  StepOutOfLoop,
  StepToNextInput, 
  StepToNextOutput
};

struct BrainFunk::Private
{
  Machine* machine;
  TapeModel* tapeModel;
  ProgramHighlighter* highlighter;

  IOHandler* userHandler;
  IOHandler* handler;

  int selectedChallenge;
  int passes;

  bool inProgress;
  int timer;
  int breakPos;
  QChar breakChar;

  QString fileName;
  bool changed;

  Private() :
    selectedChallenge(-1),
    passes(0),
    inProgress(false),
    timer(-1),
    changed(false)
  {
  }
};

BrainFunk::BrainFunk(QWidget* parent) :
  QMainWindow(parent)
{
  d = new Private();

  p = new Ui_BrainFunk();
  p->setupUi(this);

  p->widgetChallenge->setVisible(false);

  d->handler = d->userHandler = new NumericHandler(this);

  const int startingTapeSize = 1000;
  d->machine = new Machine(startingTapeSize);
  d->machine->setIOHandler(d->handler);

  d->tapeModel = new TapeModel(d->machine);
  p->tableTape->setModel(d->tapeModel);
  p->tableTape->horizontalHeader()->setResizeMode(QHeaderView::Fixed);
  p->tableTape->horizontalHeader()->setDefaultSectionSize(8 * fontMetrics().width('M'));

  d->highlighter = new ProgramHighlighter(p->editProgram->document());

  QFont monospace("Monospace", 14);
  monospace.setStyleHint(QFont::TypeWriter);
  p->editProgram->setFont(monospace);
  connect(p->editProgram, SIGNAL(modificationChanged(bool)), SLOT(changed()));

  QMenu *stepMenu = new QMenu();
  stepMenu->addAction(tr("10"))->setData(Step10);
  stepMenu->addAction(tr("100"))->setData(Step100);
  stepMenu->addAction(tr("1000"))->setData(Step1000);
  stepMenu->addAction(tr("out of loop"))->setData(StepOutOfLoop);
  stepMenu->addAction(tr("to next input"))->setData(StepToNextInput);
  stepMenu->addAction(tr("to next output"))->setData(StepToNextOutput);
  p->actionStepTo->setMenu(stepMenu);
  connect(stepMenu, SIGNAL(triggered(QAction*)), SLOT(step(QAction*)));

  ChallengeModel* model = new ChallengeModel(this);
  p->tableChallenge->setModel(model);
  p->tableChallenge->horizontalHeader()->setResizeMode(0, QHeaderView::Fixed);
  p->tableChallenge->horizontalHeader()->setResizeMode(1, QHeaderView::Stretch);
  p->tableChallenge->horizontalHeader()->setResizeMode(2, QHeaderView::Fixed);

  p->tableChallenge->horizontalHeader()->resizeSection(0, 24);
  p->tableChallenge->horizontalHeader()->resizeSection(2, 32);

  p->tableChallenge->verticalHeader()->setResizeMode(QHeaderView::Fixed);

  connect(p->tableChallenge->selectionModel(), SIGNAL(currentChanged(const QModelIndex&, const QModelIndex&)), SLOT(selectChallenge(const QModelIndex&)));

  connect(p->actionIndex, SIGNAL(triggered()), SLOT(helpIndex()));
  connect(p->actionAbout, SIGNAL(triggered()), SLOT(helpAbout()));
  connect(p->actionAboutQt, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
}

BrainFunk::~BrainFunk()
{
  delete d;
}

QPlainTextEdit* BrainFunk::editInput() const
{
  return p->editInput;
}

QPlainTextEdit* BrainFunk::editOutput() const
{
  return p->editOutput;
}

int BrainFunk::length() const
{
  QString program = p->editProgram->toPlainText();
  int length = 0;
  for (int n = 0; n < program.count(); ++n)
  {
    switch (program.at(n).toAscii())
    {
    case '+': case '-':
    case '>': case '<':
    case '[': case ']':
    case ',': case '.':
      ++length;
    default:
      break;
    }
  }

  return length;
}

bool BrainFunk::saveChanges()
{
  if (!d->changed)
    return true;

  QMessageBox::StandardButton result = QMessageBox::question(this, tr("Save changes?"), tr("Do you want to save changes to your existing program?"), QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel, QMessageBox::Yes);
  if (result == QMessageBox::Cancel)
    return false;

  if (result == QMessageBox::Yes)
  {
    QString target = d->fileName;
    if (target.isEmpty())
    {
      target = QFileDialog::getSaveFileName(this, tr("Save..."), QString(), QString("*.bf"));
      if (target.isEmpty())
        return false;
    }

    d->fileName = target;
    save();
  }

  return true;
}

void BrainFunk::save()
{
  QFile file(d->fileName);
  if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
    return;

  QTextStream stream(&file);
  stream << p->editProgram->toPlainText();
  unchanged();
}

void BrainFunk::open()
{
  QFile file(d->fileName);
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    return;

  QTextStream stream(&file);
  QStringList program;
  while (!stream.atEnd())
    program += stream.readLine();
  
  p->editProgram->setPlainText(program.join("\n"));
  unchanged();
}

void BrainFunk::changed()
{
  if (d->changed)
    return;

  d->changed = true;
  if (d->fileName.isEmpty())
    setWindowTitle(tr("BrainFunk - unsaved*"));
  else
    setWindowTitle(tr("BrainFunk - %1*").arg(d->fileName));
}

void BrainFunk::unchanged()
{
  if (!d->changed)
    return;

  p->editProgram->document()->setModified(false);
  d->changed = false;
  if (d->fileName.isEmpty())
    setWindowTitle(tr("BrainFunk"));
  else
    setWindowTitle(tr("BrainFunk - %1").arg(d->fileName));
}

void BrainFunk::on_actionNew_triggered()
{
  if (!saveChanges())
    return;

  p->editProgram->setPlainText(QString());
  d->fileName = QString();
  unchanged();
}

void BrainFunk::on_actionOpen_triggered()
{
  if (!saveChanges())
    return;

  QString target = QFileDialog::getOpenFileName(this, tr("Open..."), QString(), QString("*.bf"));
  if (target.isEmpty())
    return;

  d->fileName = target;
  open();
}

void BrainFunk::on_actionSave_triggered()
{
  if (d->fileName.isEmpty())
  {
    QString target = QFileDialog::getSaveFileName(this, tr("Save..."), QString(), QString("*.bf"));
    if (target.isEmpty())
        return;

    d->fileName = target;
  }

  save();
}

void BrainFunk::on_actionSaveAs_triggered()
{
  QString target = QFileDialog::getSaveFileName(this, tr("Save..."), QString(), QString("*.bf"));
  if (target.isEmpty())
    return;
  
  d->fileName = target;
  save();
}

void BrainFunk::on_actionLength_triggered()
{
  QMessageBox::information(this, tr("Length"), tr("Your program contains %1 commands.").arg(length()));
}

void BrainFunk::on_actionTapeLength_triggered()
{
  int length = d->machine->tapeSize();
  bool result;
  length = QInputDialog::getInt(this, "Enter tape length...", "Length:", length, 1, 50000, 1, &result);

  if (result)
  {
    d->machine->setTapeSize(length);
    d->tapeModel->tapeChanged();
  }
}

void BrainFunk::on_actionMaximumValue_triggered()
{
  int value = d->machine->maximumValue();
  bool result;
  value = QInputDialog::getInt(this, "Enter maximum value...", "Value:", value, 1, 2000000000, 1, &result); // I just stuck 2 billion in there lol

  if (result)
    d->machine->setMaximumValue(value );
}

void BrainFunk::on_actionAlphabet_triggered(bool alphabet)
{
  if (alphabet)
    d->handler = d->userHandler = new AlphaHandler(this);
  else
    d->handler = d->userHandler = new NumericHandler(this);

  d->machine->setIOHandler(d->handler);
}

void BrainFunk::on_actionRun_triggered()
{
  if (!d->inProgress)
    prepare();

  if (d->timer == -1)
    start();
  else
  {
    stop();
    updateGUI();
  }
}

void BrainFunk::on_actionStep_triggered()
{
  step(Step1);
}

void BrainFunk::on_actionHalt_triggered()
{
  halt();
  updateGUI();
}

void BrainFunk::on_buttonTest_clicked()
{
  if (d->selectedChallenge == -1)
    return;

  if (d->inProgress)
  {
    halt();
    return;
  }

  // Start testing the current challenge!!
  Challenge& current = global::challenges[d->selectedChallenge];

  // Enforce program length
  if (current.programLength && length() > current.programLength)
  {
    QMessageBox::warning(this, tr("Program length"), tr("Your program is %1 commands. This task requires a program of %2 or less commands.").arg(length()).arg(current.programLength));
    return;
  }

  d->passes = current.data.count();

  d->machine->setMaximumValue(Machine::DefaultMaximum);
  d->machine->setTapeSize(current.tapeLength);
  d->tapeModel->tapeChanged();

  d->handler = current.createHandler(this);
  Q_ASSERT(d->handler);
  d->machine->setIOHandler(d->handler);

  p->toolBar->setEnabled(false);
  p->tableChallenge->setEnabled(false);

  p->buttonTest->setText(tr("Cancel"));

  prepare();
  start();
}

void BrainFunk::helpIndex()
{
  QMessageBox::critical(this, tr("Serious warning"), tr("Nobody can help you."));
}

void BrainFunk::helpAbout()
{
  QMessageBox::information(this, tr("About"), tr("This is a brainfuck interpreter by Shambles Software.\n\nhttp://www.shamblessoftware.co.uk/"));
}

void BrainFunk::selectChallenge(const QModelIndex& index)
{
  int row = index.row();

  if (row >= global::challenges.count())
    return;

  d->selectedChallenge = row;

  p->editChallenge->setPlainText(global::challenges[row].description);
  p->buttonTest->setEnabled(true);
  d->machine->setMaximumValue(Machine::DefaultMaximum);

  int tape = global::challenges[row].tapeLength;
  d->machine->setTapeSize(tape);

  int program = global::challenges[row].programLength;

  QString requirements;
  if (program)
    requirements = tr("Tape length %1, program length %2").arg(tape).arg(program);
  else
    requirements = tr("Tape length %1").arg(tape);
  p->labelRequirements->setText(requirements);

  d->tapeModel->tapeChanged();
}

void BrainFunk::step(QAction* action)
{
  int stepcode = action->data().toInt();
  step(stepcode);
}

void BrainFunk::timerEvent(QTimerEvent*)
{
  run(1000);
}

void BrainFunk::closeEvent(QCloseEvent* close)
{
  if (d->inProgress)
    halt();

  if (!saveChanges())
    close->ignore();
}

void BrainFunk::continueChallenge()
{
  // A challenge test has just completed, let's continue
  if (--d->passes == 0)
  {
    // Reached the pass count, win
    updateGUI();
    QMessageBox::information(this, tr("Pass"), tr("Your program has passed the challenge!"));
    endChallenge();
    global::setChallengeComplete(d->selectedChallenge);
  }
  else
  {
    prepare();
    start();
  }
}

void BrainFunk::endChallenge()
{
  d->machine->setIOHandler(d->userHandler);

  delete d->handler;
  d->handler = d->userHandler;

  d->passes = 0;

  p->toolBar->setEnabled(true);
  p->tableChallenge->setEnabled(true);

  p->buttonTest->setText(tr("Test"));
}

void BrainFunk::prepare()
{
  p->actionHalt->setEnabled(true);
  p->buttonTest->setEnabled(d->passes > 0);
  p->menuBar->setEnabled(false);

  p->editInput->setReadOnly(true);
  p->editProgram->setReadOnly(true);

  QString program = p->editProgram->toPlainText();
  for (int n = 0; n < program.count(); ++n)
  {
    QChar p = program.at(n);
    switch (p.toAscii())
    {
    default:
      program[n] = QChar(' ');
    case '+': case '-':
    case '>': case '<':
    case '[': case ']':
    case ',': case '.':
    case '@':
      break;
    }
  }

  d->machine->setProgram(program);
  d->machine->reset();

  d->breakPos = -1;
  d->breakChar = d->passes ? QChar() : QChar('@');

  d->inProgress = true;

  d->handler->reset();
  updateGUI();
}

void BrainFunk::run(int steps)
{
  bool freeze = false;
  while (!freeze && steps--)
  {
    freeze = d->machine->step();
    readOutput();

    if (d->breakPos != -1 && d->machine->programPosition() == d->breakPos)
      freeze = true;

    if (d->breakChar != QChar() && d->machine->programValue() == d->breakChar)
      freeze = true;
  }

  if (freeze)
  {
    d->breakChar = QChar('@');
    d->breakPos = -1;
    stop();
  }

  updateGUI();

  if (d->machine->haltCode() != Machine::HaltNone)
    halt();
}

void BrainFunk::step(int mode)
{
  if (!d->inProgress)
    prepare();

  stop();

  switch (mode)
  {
  default:
    break;
  case Step1:
    run(1);
    break;
  case Step10:
    run(10);
    break;
  case Step100:
    run(100);
    break;
  case Step1000:
    run(1000);
    break;
  case StepOutOfLoop:
    d->breakChar = QChar();
    d->breakPos = d->machine->findLoopExit(d->machine->programPosition());
    start();
    break;
  case StepToNextInput:
    d->breakChar = QChar(',');
    start();
    break;
  case StepToNextOutput:
    d->breakChar = QChar('.');
    start();
    break;
  }
}

void BrainFunk::start()
{
  stop();
  d->timer = startTimer(0);
}

void BrainFunk::stop()
{
  if (d->timer == -1)
    return;

  killTimer(d->timer);
  d->timer = -1;
}

void BrainFunk::halt()
{
  stop();
  
  p->actionHalt->setEnabled(false);
  p->buttonTest->setEnabled(true);
  p->menuBar->setEnabled(true);

  p->editInput->setReadOnly(false);
  p->editProgram->setReadOnly(false);

  d->inProgress = false;

  switch (d->machine->haltCode())
  {
  case Machine::HaltNormal:
    if (d->passes)
      continueChallenge();
    return;

  case Machine::HaltOverLeft:
    QMessageBox::critical(this, tr("Crash"), tr("Invalid move over left side of tape!"));
    break;
  case Machine::HaltOverRight:
    QMessageBox::critical(this, tr("Crash"), tr("Invalid move over right side of tape!"));
    break;
  case Machine::HaltBadInput:
    QMessageBox::critical(this, tr("Crash"), tr("Bad input!"));
    break;
  case Machine::HaltBadOutput:
    QMessageBox::critical(this, tr("Crash"), tr("Bad output!"));
    break;
  case Machine::HaltUnmatchedLoop:
    QMessageBox::critical(this, tr("Crash"), tr("Unmatched loop!"));
    break;
  case Machine::HaltUnfinishedResult:
    QMessageBox::critical(this, tr("Crash"), tr("The program terminated without delivering the full result!"));
    break;
  default:
  case Machine::HaltUnknown:
    QMessageBox::critical(this, tr("Crash"), tr("Unknown crash! How odd."));
    break;
  case Machine::HaltNone:
    break;
  }

  if (d->passes)
  {
    QMessageBox::warning(this, tr("Fail"), tr("The program failed the challenge."));
    endChallenge();
  }
}

void BrainFunk::updateGUI()
{
  p->tableTape->reset();
  p->tableTape->scrollTo(d->tapeModel->positionIndex(), QAbstractItemView::PositionAtCenter);

  d->handler->refresh();

  if (d->inProgress || d->machine->haltCode() != Machine::HaltNormal)
    d->highlighter->setExecutionPosition(d->machine->programPosition());
  else
    d->highlighter->setExecutionPosition(-1);
}

void BrainFunk::readOutput()
{
}


