#include "ChallengeHandlers.h"

#include "BrainFunk.h"

#include <QPlainTextEdit>
#include <QStringList>

struct NormalTestHandler::Private
{
  QPlainTextEdit* editInput;
  QPlainTextEdit* editOutput;

  QStringList data;
  int pass;

  QList<int> input;
  QList<int> output;

  Private() :
    pass(0)
  {
  }
};

NormalTestHandler::NormalTestHandler(BrainFunk* bf, const QStringList& data) :
  IOHandler()
{
  d = new Private();

  d->editInput = bf->editInput();
  d->editOutput = bf->editOutput();

  d->data = data;
}

NormalTestHandler::~NormalTestHandler()
{
  delete d;
}

void NormalTestHandler::reset()
{
  d->input.clear();
  d->output.clear();

  QStringList parts = d->data[d->pass++].split(" => ", QString::KeepEmptyParts);

  if (parts.count() != 2)
    return;

  if (!parts.first().isEmpty())
  {
    d->editInput->setPlainText(QObject::tr("Test input: %1").arg(parts.first()));

    QStringList inputText = parts.first().split(",");
    foreach (const QString& text, inputText)
      d->input.append(text.toInt());
  }
  else
    d->editInput->setPlainText(QObject::tr("No test input."));

  d->editOutput->setPlainText(QObject::tr("Required output: %1").arg(parts.last()));
  
  if (!parts.last().isEmpty())
  {
    QStringList outputText = parts.last().split(",");
    foreach (const QString& text, outputText)
      d->output.append(text.toInt());
  }
}

void NormalTestHandler::refresh()
{
}

bool NormalTestHandler::hasInput()
{
  return !d->input.isEmpty();
}

int NormalTestHandler::input()
{
  return d->input.takeFirst();
}

bool NormalTestHandler::canOutput(int output)
{
  if (d->output.isEmpty())
    return false;

  return d->output.takeFirst() == output;
}

bool NormalTestHandler::canFinish()
{
  return d->input.isEmpty() && d->output.isEmpty();
}

// ***************

struct AlphaTestHandler::Private
{
  QPlainTextEdit* editInput;
  QPlainTextEdit* editOutput;

  QStringList data;
  int pass;

  QString input, output;
  int inputPosition, outputPosition;

  Private() :
    pass(0)
  {
  }
};

AlphaTestHandler::AlphaTestHandler(BrainFunk* bf, const QStringList& data) :
  IOHandler()
{
  d = new Private();

  d->editInput = bf->editInput();
  d->editOutput = bf->editOutput();

  d->data = data;
}

AlphaTestHandler::~AlphaTestHandler()
{
  delete d;
}

void AlphaTestHandler::reset()
{
  d->inputPosition = d->outputPosition = 0;
  QStringList parts = d->data[d->pass++].split(" => ", QString::KeepEmptyParts);

  if (parts.count() != 2)
    return;

  d->input = parts.first();
  if (!d->input.isEmpty())
    d->input += QChar('\n'); // Alpha inputs are null terminated when they exist
  d->input = d->input.replace(QString("\\n"), QString("\n"));

  d->editInput->setPlainText(d->input);

  d->output = parts.last();
  d->editOutput->setPlainText(QObject::tr("Required output: %1").arg(d->output));
}

void AlphaTestHandler::refresh()
{
}

bool AlphaTestHandler::hasInput()
{
  return d->inputPosition < d->input.length();
}

int AlphaTestHandler::input()
{
  QChar c = d->input.at(d->inputPosition++);

  if (c == QChar('\n'))
    return 0;

  return c.toAscii() - 'a' + 1;
}

bool AlphaTestHandler::canOutput(int output)
{
  if (d->outputPosition == d->output.length())
    return false;

  QChar c = d->output.at(d->outputPosition++);
  if (output == 0)
    return c == QChar('\n');
  return output == c.toAscii() - 'a' + 1;
}

bool AlphaTestHandler::canFinish()
{
  return d->inputPosition == d->input.length() && d->outputPosition == d->output.length();
}

struct NimTestHandler::Private
{
  QPlainTextEdit* editInput;
  QPlainTextEdit* editOutput;
  
  QStringList history;

  int pile[3];
  int selectedPile;
  bool pilesEmpty() const { return !(pile[0] + pile[1] + pile[2]); }
  
  enum Stage
  {
    DeclarePileOne,
    DeclarePileTwo,
    DeclarePileThree,
    ComputerPile,
    ComputerCount,
    UserPile,
    UserCount
  };

  Stage stage;
};

NimTestHandler::NimTestHandler(BrainFunk* bf) :
  IOHandler()
{
  d = new Private();

  d->editInput = bf->editInput();
  d->editOutput = bf->editOutput();
}

NimTestHandler::~NimTestHandler()
{
  delete d;
}

void NimTestHandler::reset()
{
  d->stage = Private::DeclarePileOne;

  d->pile[0] = 1 + rand() % 30;
  d->pile[1] = 1 + rand() % 30;

  // Weight the piles in favour of a nim-sum based AI player
  while ( !(d->pile[2] = d->pile[0] ^ d->pile[1] ^ (1 + rand() % 30)));

  d->history.clear();
  for (int n = 0; n < 3; ++n)
    d->history += QString::number(d->pile[n]);
}

void NimTestHandler::refresh()
{
  QString piles = QString("Piles: %1 %2 %3").arg(d->pile[0]).arg(d->pile[1]).arg(d->pile[2]);
  d->editOutput->setPlainText(piles);

  d->editInput->setPlainText(QString("Computer play: %1").arg(d->history.join(",")));
}

bool NimTestHandler::hasInput()
{
  switch(d->stage)
  {
  case Private::UserPile: case Private::UserCount:
    return false;
  case Private::ComputerPile:
    return !d->pilesEmpty();
  default:
    return true;
  }
}

int NimTestHandler::input()
{
  switch(d->stage)
  {
  case Private::DeclarePileOne:
    d->stage = Private::DeclarePileTwo;
    return d->pile[0];
  case Private::DeclarePileTwo:
    d->stage = Private::DeclarePileThree;
    return d->pile[1];
  case Private::DeclarePileThree:
    d->stage = Private::UserPile;
    return d->pile[2];
  case Private::ComputerPile:
    {
      // Bad nim AI goes here
      // First select a pile with >1 element if possible
      QList<int> options;
      for (int n = 0; n < 3; ++n)
        if (d->pile[n] > 1)
          options += n;

      if (options.count())
        d->selectedPile = options[rand() % options.count()];
      else
      {
        // Otherwise select a non-zero pile
        for (int n = 0; n < 3; ++n)
          if (d->pile[n] > 0)
            options += n;
        
        d->selectedPile = options[rand() % options.count()];
      }

      d->stage = Private::ComputerCount;
      d->history += QString::number(d->selectedPile + 1);
      return d->selectedPile + 1;
    }
  case Private::ComputerCount:
    {
      // More bad AI!!
      int count = 1; // Have to take at least something
      if (d->pile[d->selectedPile] > 8)
        count = 1 + rand() % 5; // reduce large piles a little
      if (d->pile[d->selectedPile] > 1)
        count = d->pile[d->selectedPile] - 1; // reduce medium pile to 1
      d->stage = Private::UserPile;
      d->pile[d->selectedPile] -= count;
      d->history += QString::number(count);
      return count;
    }
  default:
    Q_ASSERT(false);
    return 0;
  }
}

bool NimTestHandler::canOutput(int output)
{
  if (d->stage == Private::UserPile)
  {
    d->stage = Private::UserCount;
     if (output == 0 || output > 3)
       return false;
     if (d->pile[output - 1] == 0)
       return false;
     d->selectedPile = output - 1;
  }
  else if (d->stage == Private::UserCount)
  {
    d->stage = Private::ComputerPile;
    if (d->pile[d->selectedPile] < output)
      return false;
    d->pile[d->selectedPile] -= output;
  }
  else
    return false;
  
  return true;
}

bool NimTestHandler::canFinish()
{
  return d->pilesEmpty();
}
