#include "IOHandler.h"

#include "BrainFunk.h"

#include <QPlainTextEdit>
#include <QInputDialog>

#include <climits>

// ***** Default handler *****

IOHandler::IOHandler()
{
}

IOHandler::~IOHandler()
{
}

bool IOHandler::hasInput()
{
  return false;
}

int IOHandler::input()
{
  return 0;
}

bool IOHandler::canOutput(int)
{
  return false;
}

bool IOHandler::canFinish()
{
  return true;
}

// ***** Numeric Handler *****
// Used when not in text mode

struct NumericHandler::Private
{
  QPlainTextEdit* input;
  QPlainTextEdit* output;
  
  QString inputString;
  int inputPosition;
  
  int inputValue;

  QStringList outputList;
};

NumericHandler::NumericHandler(BrainFunk* bf)
{
  d = new Private();
  
  d->input = bf->editInput();
  d->output = bf->editOutput();
}

NumericHandler::~NumericHandler()
{
  delete d;
}

void NumericHandler::reset()
{
  d->inputString = d->input->toPlainText();
  d->inputPosition = 0;

  d->outputList.clear();
  refresh();
}

void NumericHandler::refresh()
{
  QString result = d->outputList.join(", ");

  if (!result.isEmpty())
  {
    d->outputList.clear();
    d->outputList += result;
  }
  
  d->output->setPlainText(result);
}

bool NumericHandler::hasInput()
{
  while (d->inputPosition < d->inputString.length() && !d->inputString.at(d->inputPosition).isDigit())
    ++d->inputPosition;

  if (d->inputPosition < d->inputString.length())
  {
    // read from input to inputValue
    int end = d->inputPosition + 1;
    while (end < d->inputString.length() && d->inputString.at(end).isDigit())
      ++end;
    QString mid = d->inputString.mid(d->inputPosition, end - d->inputPosition);
    d->inputValue = mid.toInt();
    d->inputPosition = end;
    return true;
  }
    
  // query user
  bool result;
  d->inputValue = QInputDialog::getInt(d->input, "Enter input...", "Input:", 0, 0, INT_MAX, 1, &result);
  return result;
}

int NumericHandler::input()
{
  return d->inputValue;
}

bool NumericHandler::canOutput(int output)
{
  d->outputList += QString::number(output);
  return true;
}

// ***** Alpha Handler *****
// Used when in text mode

struct AlphaHandler::Private
{
  QPlainTextEdit* input;
  QPlainTextEdit* output;

  QString inputString;
  int inputPosition;

  int inputValue;

  QString outputString;
};

AlphaHandler::AlphaHandler(BrainFunk* bf)
{
  d = new Private();

  d->input = bf->editInput();
  d->output = bf->editOutput();
}

AlphaHandler::~AlphaHandler()
{
  delete d;
}

void AlphaHandler::reset()
{
  d->inputString = d->input->toPlainText();
  d->inputPosition = 0;

  d->outputString = QString();
  refresh();
}

void AlphaHandler::refresh()
{
  d->output->setPlainText(d->outputString);
}

bool AlphaHandler::hasInput()
{
  while (d->inputPosition < d->inputString.length() && !(d->inputString.at(d->inputPosition).isLetter() || d->inputString.at(d->inputPosition) == QChar('\n')))
    ++d->inputPosition;

  if (d->inputPosition >= d->inputString.length())
  {
    // query user
    bool result;
    QString data = QInputDialog::getText(d->input, "Enter input...", "Input:", QLineEdit::Normal, "", &result);
    
    if (!result)
      return false;
    
    d->inputString += data + QChar('\n');
  }

  while (!(d->inputString.at(d->inputPosition).isLetter() || d->inputString.at(d->inputPosition) == QChar('\n')))
    ++d->inputPosition;
  
  QChar letter = d->inputString.at(d->inputPosition++).toLower();
  d->inputValue = (letter == QChar('\n') ? 0 : (1 + letter.toAscii() - 'a'));
  return true;
}

int AlphaHandler::input()
{
  return d->inputValue;
}

bool AlphaHandler::canOutput(int output)
{
  if (output <= 26)
  {
    if (output == 0)
      d->outputString += QChar::LineSeparator;
    else
      d->outputString += QChar('a' + output - 1);
    return true;
  }

  return false;
}
