#include "Machine.h"

#include <QList>
#include <QVector>

#include "IOHandler.h"

struct Machine::Private
{
  QString program;
  int programPointer;
  QList<int> programPointerStack;
  
  QVector<int> tape;
  int tapePointer;
  
  int maximum;
  
  IOHandler* handler;

  bool running;
  HaltCode haltCode;
  
  Private() :
    maximum(33713),
    handler(NULL)
  {
  }
};

Machine::Machine()
{
  d = new Private();
  setTapeSize(20000);
}

Machine::Machine(int tapeSize)
{
  d = new Private();
  setTapeSize(tapeSize);
}

Machine::~Machine()
{
  delete d;
}

void Machine::setIOHandler(IOHandler* handler)
{
  d->handler = handler;
}

int Machine::tapeSize() const
{
  return d->tape.count();
}

void Machine::setTapeSize(int size)
{
  d->tape.resize(size);
  clear();
}
  
int Machine::maximumValue() const
{
  return d->maximum;
}

void Machine::setMaximumValue(int max)
{
  Q_ASSERT(max >= 1);
  d->maximum = max;
}
  
void Machine::setProgram(const QString& program)
{
  d->program = program;
}

void Machine::reset()
{
  clear();
  d->running = true;
}

void Machine::clear()
{
  d->programPointer = -1;
  d->programPointerStack.clear();

  d->tape.fill(0);
  d->tapePointer = 0;

  d->running = false;

  d->haltCode = HaltNone;
}

bool Machine::step() // Returns true if this machine stops playback
{
  QChar code;

  code = readCode();
  
  if (!d->running)
    return true;
  
  switch(code.toAscii())
  {
  case '+':
    if (++d->tape[d->tapePointer] > d->maximum)
      d->tape[d->tapePointer] = 0;
    break;

  case '-':
    if (d->tape[d->tapePointer] == 0)
      d->tape[d->tapePointer] = d->maximum;
    else
      --d->tape[d->tapePointer];
    break;

  case '>':
    ++d->tapePointer;
    if (d->tapePointer >= d->tape.count())
      halt(HaltOverRight);
    break;

  case '<':
    --d->tapePointer;
    if (d->tapePointer < 0)
      halt(HaltOverLeft);
    break;

  case '[':
    if (d->tape[d->tapePointer] == 0)
    {
      int target = findLoopExit(d->programPointer);
      if (target == -1)
        halt(HaltUnmatchedLoop);
      else
        d->programPointer = target;
    }
    else
      d->programPointerStack += d->programPointer;
    break;

  case ']':
    if (d->programPointerStack.isEmpty())
      halt(HaltUnmatchedLoop);
    else
    {
      if (d->tape[d->tapePointer] != 0)
        d->programPointer = d->programPointerStack.last();
      else
        d->programPointerStack.removeLast();
    }
    break;

  case ',':
    if (!d->handler->hasInput())
      halt(HaltBadInput);
    else
    {
      int input = d->handler->input();
      if (input < 0 || input > d->maximum)
        halt(HaltBadInput);
      else
        d->tape[d->tapePointer] = input;
    }
    break;

  case '.':
    if (!d->handler->canOutput(d->tape[d->tapePointer]))
      halt(HaltBadOutput);
    break;

  default:
    break;
  }

  return !d->running;
}
  
Machine::HaltCode Machine::haltCode() const
{
  return d->haltCode;
}
  
int Machine::programPosition() const
{
  return d->programPointer;
}

QChar Machine::programValue() const
{
  return d->program[d->programPointer];
}

int Machine::tapePosition() const
{
  return d->tapePointer;
}

int Machine::tapeValue(int position) const
{
  if (position < 0 || position >= d->tape.count())
    return 0;
  return d->tape[position];
}

int Machine::findLoopExit(int position) const
{
  int level = 1;
  while (++position < d->program.length())
  {
    if (d->program[position] == '[')
      ++level;
    if (d->program[position] == ']')
      --level;
    if (level == 0)
      return position;
  }

  return -1;
}

QChar Machine::readCode()
{
  ++d->programPointer;
  while (d->programPointer < d->program.length())
  {
    if (d->program[d->programPointer] != ' ')
      return d->program[d->programPointer];
    ++d->programPointer;
  }
  
  if (!d->programPointerStack.isEmpty())
    halt(HaltUnmatchedLoop);
  else if (!d->handler->canFinish())
    halt(HaltUnfinishedResult);
  else
    halt(HaltNormal);

  return QChar();
}

void Machine::halt(HaltCode code)
{
  d->haltCode = code;
  d->running = false;
}
