#include "ProgramHighlighter.h"

#include <QTextDocument>
#include <QTextBlock>

struct ProgramHighlighter::Private
{
  QTextCharFormat moveFormat;
  QTextCharFormat moveTallyFormat;
  QTextCharFormat changeFormat;
  QTextCharFormat changeTallyFormat;
  QTextCharFormat controlFormat;
  QTextCharFormat IOFormat;
  QTextCharFormat executionFormat;

  QTextBlock executionBlock;
  int blockPosition;
  
  Private() :
    blockPosition(-1)
  {
  }
};

ProgramHighlighter::ProgramHighlighter(QTextDocument* parent) :
  QSyntaxHighlighter(parent)
{
  d = new Private();

  d->moveFormat.setForeground(QColor(0xC0, 0, 0));
  d->moveTallyFormat.setForeground(d->moveFormat.foreground());
  d->moveTallyFormat.setUnderlineStyle(QTextCharFormat::DotLine);

  d->changeFormat.setForeground(QColor(0, 0x80, 0xFF));
  d->changeTallyFormat.setForeground(d->changeFormat.foreground());
  d->changeTallyFormat.setUnderlineStyle(QTextCharFormat::DotLine);

  d->controlFormat.setForeground(QColor(0x80, 0, 0xC0));

  d->IOFormat.setForeground(QColor(0, 0x80, 0));
  d->IOFormat.setFontWeight(QFont::Bold);

  d->executionFormat.setForeground(QColor(0xFF, 0xFF, 0xFF));
  d->executionFormat.setBackground(QColor(0xFF, 0x80, 0));
  d->executionFormat.setFontWeight(QFont::Black);
}

ProgramHighlighter::~ProgramHighlighter()
{
  delete d;
}

void ProgramHighlighter::setExecutionPosition(int position)
{
  // Clean up old highlight
  if (d->executionBlock.isValid())
    rehighlightBlock(d->executionBlock);

  if (position == -1)
  {
    d->executionBlock = QTextBlock();
    return;
  }

  d->executionBlock = document()->findBlock(position);
  d->blockPosition = position - d->executionBlock.position();
  rehighlightBlock(d->executionBlock);
  d->blockPosition = -1;
}

void ProgramHighlighter::highlightBlock(const QString& text)
{
  HighlightCode prev, now = NoHighlight;
  int start = -1;
  for (int index = 0;  index < text.length(); ++index)
  {
    prev = now;
    now = code(text.at(index));

    if (now != prev)
    {
      if (prev != NoHighlight)
        applyCode(prev, start, index - start);
      start = index;
    }
  }

  if (now != NoHighlight)
    applyCode(now, start, text.length() - start);

  if (d->blockPosition > -1)
    setFormat(d->blockPosition, 1, d->executionFormat);
}

void ProgramHighlighter::applyCode(HighlightCode what, int from, int length)
{
  switch (what)
  {
  case HighlightMove:
    applyTally(d->moveFormat, d->moveTallyFormat, from, length);
    return;
  case HighlightChange:
    applyTally(d->changeFormat, d->changeTallyFormat, from, length);
    return;
  case HighlightControl:
    setFormat(from, length, d->controlFormat);
    break;
  case HighlightIO:
    setFormat(from, length, d->IOFormat);
    break;
  default:
    break;
  }
}

void ProgramHighlighter::applyTally(QTextCharFormat& regular, QTextCharFormat& tally, int from, int length)
{
  while (length)
  {
    if (length >= 5)
    {
      setFormat(from, 4, regular);
      setFormat(from + 4, 1, tally);
      from += 5;
      length -= 5;
    }
    else
    {
      setFormat(from, length, regular);
      break;
    }
  }
}

ProgramHighlighter::HighlightCode ProgramHighlighter::code(QChar letter)
{
  switch (letter.toAscii())
  {
  case '<': case '>':
    return HighlightMove;
  case '+': case '-':
    return HighlightChange;
  case '[': case ']': case '@':
    return HighlightControl;
  case ',': case '.':
    return HighlightIO;
  default:
    return NoHighlight;
  }
}
